from orquestra.quantum.circuits import CNOT, H, Circuit, X, Z

from orquestra.quantum.symbolic_simulator import SymbolicSimulator

from icecream import ic

teleport_circuit = Circuit()

# prepare entanglement
teleport_circuit += H(0)
teleport_circuit += CNOT(0, 1)

# alice encodes bits into her qubit
# comment and uncomment these lines to send 00, 01, 10, and 11
teleport_circuit += X(0)
# teleport_circuit += Z(0)

# bob decodes the qubits
teleport_circuit += CNOT(0, 1)
teleport_circuit += H(0)

# bob runs and measures the qubits
backend = SymbolicSimulator()
measurements = backend.run_circuit_and_measure(teleport_circuit, 1)
ic(measurements.get_counts())



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
from multiprocessing.sharedctypes import Value
import subprocess
import sys
import pathlib
import json
import os
from glob import glob
from pathlib import Path
import shutil
from distutils.dir_util import copy_tree
from tempfile import tempdir


repos_folder = os.path.abspath("repos")
TMP_MONOREPO_NAME = "tmp_monorepo"
currentfolder = Path(os.path.abspath("."))
temp_repo_folder = currentfolder / f"{TMP_MONOREPO_NAME}"

temp_repo_folder = str(temp_repo_folder)
sys.path.insert(0, temp_repo_folder)


def build_mono_repo(temp_repo_folder):

    subprocess.call(
        ["mkdir", "-p", temp_repo_folder],
        shell = True
    )  # make sure temporary folder exists

    # scan the manifest for the location of the root code directory for each repo
    with open(os.path.join(repos_folder, "manifest.json")) as fp:
        manifest = json.load(fp)

    for name, details in manifest["repos"].items():
        for path in details.get("autodoc", []):
            source_dir = str(Path(name) / path)
            print(f"source_dir: {source_dir} to tmp folder: {temp_repo_folder}")
            copy_tree(
                os.path.join(repos_folder, source_dir), temp_repo_folder
            )  # copy all folders to docs/tmp/repo/*


try:
    shutil.rmtree(temp_repo_folder)
except FileNotFoundError:
    print(f"no dir to remove {temp_repo_folder}")

build_mono_repo(temp_repo_folder)
# -- Project information -----------------------------------------------------

project = "Orquestra Core Docs"
copyright = "2022, Zapata Computing, Inc"

# The full version, including alpha/beta/rc tags
release = "0.2.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.doctest",
    "sphinx_design",
    # "sphinx.ext.graphviz",
    # "sphinxcontrib.youtube",
    # "sphinxcontrib.autoprogram",
    # "sphinxcontrib.confluencebuilder",
    "sphinx_copybutton",
    "sphinxemoji.sphinxemoji",
    "sphinx_togglebutton",
    "autoapi.extension",
]
source_suffix = {
    ".rst": "restructuredtext",
    # '.txt': 'restructuredtext',
    # '.md': 'markdown',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "venv",
    "repos",
    "subtrees",
    "developer",
    "investigator",
    "src",
    "tests",
    "build",
    ".venv",
    ".github",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "sphinx_book_theme"
html_theme_options = {
    "use_download_button": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# this gets around some kind of bug-like behavior in autoapi where it doesn't create
# the entries for the index.rst unless there is an __init__.py despite
# setting autoapi_python_use_implicit_namespaces=True

autoapi_dirs = glob(os.path.join(temp_repo_folder, "orquestra", "*"))
autoapi_root = "api"
for dir in autoapi_dirs:
    Path(dir, "__init__.py").touch()


html_logo = "_static/orquestra.png"
html_title = "Orquestra Core Documentation"
html_favicon = "_static/favicon.ico"

# Intersphinx is a tool for creating links to multiple repos.  This should help us!
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# This is so that the args to __init__ methods are appended to the class docs
autoclass_content = "both"

# set to true to see the contents of the autoapi directory after building
autoapi_keep_files = True
autoapi_python_use_implicit_namespaces = True

autoapi_options = [
    "members",
    "show-inheritance",
    "show-module-summary",
    "imported-members",
    "undoc-members",
]


# uses a custom CSS file to override some defaults
def setup(app):
    app.add_css_file("css/autoapi.css")



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################

from orquestra.quantum.circuits import (
    Circuit,
    H,
    RZ,
    CNOT,
    RX,
    CZ,
    CustomGateDefinition,
)
from orquestra.quantum.decompositions import decompose_operations
import sympy
import numpy as np

from icecream import ic

N_QUBITS = 3

beta = sympy.symbols("beta_0:3", real=True)
gamma = sympy.symbols("gamma_0:3", real=True)
theta = sympy.Symbol("theta", real=True)

# State Preparation Circuit
state_prep_circ = Circuit([H(i) for i in range(N_QUBITS)])
unitary = state_prep_circ.to_unitary()
ic(unitary.shape)  # should be 8x8 matrix
ic(type(unitary))  # will be numpy ndarray, later see sympy matrix

# Inverting the circuit
inverse_state_prep_circ = state_prep_circ.inverse()
ic(inverse_state_prep_circ)
combined_circ = state_prep_circ + inverse_state_prep_circ
combined_unitary = combined_circ.to_unitary()
ic(np.allclose(combined_unitary, np.eye(8)))

# Problem Hamiltonian Circuit
problem_hamiltonian_circ = Circuit([CNOT(0, 1), RZ(beta[0])(1), CNOT(0, 1)])
problem_hamiltonian_circ += Circuit([CNOT(0, 2), RZ(beta[1])(2), CNOT(0, 2)])
problem_hamiltonian_circ += Circuit([CNOT(1, 2), RZ(beta[2])(2), CNOT(1, 2)])
unitary = problem_hamiltonian_circ.to_unitary()
ic(unitary.shape)  # should be 8x8 matrix
ic(type(unitary))  # sympy matrix

# Mixing Hamiltonian Circuit
mixing_circ = Circuit()
for i in range(N_QUBITS):
    mixing_circ += RX(gamma[i])(i)
ic(mixing_circ)

# Combine the circuit
qaoa_circ = state_prep_circ + problem_hamiltonian_circ + mixing_circ
ic(qaoa_circ)
ic(qaoa_circ.n_qubits)
ic(len(qaoa_circ.operations))
for op in qaoa_circ.operations:
    ic(op.gate.name, op.qubit_indices, op.params)
ic(qaoa_circ.free_symbols)

# Binding Beta Circuit
bound_beta_circ = qaoa_circ.bind({beta[0]: 0.1, beta[1]: 0.1, beta[2]: 0.1})
ic(bound_beta_circ.free_symbols)

# Binding Gammas in Circuit
bound_all_circ = bound_beta_circ.bind({gamma[0]: 0.2, gamma[1]: 0.2, gamma[2]: 0.2})
ic(bound_all_circ.free_symbols)
unitary = bound_all_circ.to_unitary()
ic(type(unitary))

# Non-parametrized mixing circuit for comparison
new_mixing_circ = Circuit([RX(0.2)(0), RX(0.2)(1), RX(0.2)(2)])
ic(new_mixing_circ == mixing_circ.bind({gamma[0]: 0.2, gamma[1]: 0.2, gamma[2]: 0.2}))

# Custom Gates
custom_matrix = sympy.Matrix(
    [
        [1.0 * sympy.exp(-sympy.I * theta / 2), 0, 0, 0],
        [0, 1.0 * sympy.exp(sympy.I * theta / 2), 0, 0],
        [0, 0, 1.0 * sympy.exp(sympy.I * theta / 2), 0],
        [0, 0, 0, 1.0 * sympy.exp(-sympy.I * theta / 2)],
    ]
)
ZZ = CustomGateDefinition(
    gate_name="ZZ", matrix=custom_matrix, params_ordering=(theta,)
)
new_problem_ham_circ = Circuit(
    [ZZ(beta[0])(0, 1), ZZ(beta[1])(0, 2), ZZ(beta[2])(1, 2)]
)
ic(new_problem_ham_circ.to_unitary() == problem_hamiltonian_circ.to_unitary())
ic(new_problem_ham_circ.collect_custom_gate_definitions())


# Decomposing Gates
class CNOTDecompositionRule:
    def production(self, operation):
        q0, q1 = operation.qubit_indices
        return [H(q1), CZ(q0, q1), H(q1)]

    def predicate(self, operation):
        return operation.gate.name == "CNOT"


decomposed_circ = Circuit(
    decompose_operations(bound_all_circ.operations, [CNOTDecompositionRule()])
)
ic(np.allclose(decomposed_circ.to_unitary(), bound_all_circ.to_unitary()))


# Exponential and power gates
ic(H.power(2).matrix == sympy.eye(2))

# exponential example
ic(sympy.simplify(H.exp.matrix))



################################################################################
# © Copyright 2022 Zapata Computing Inc.
# modified from tests/qaoa_maxcut_test.py
################################################################################
from collections import Counter

import networkx as nx
import numpy as np

from orquestra.integrations.cirq.simulator import CirqSimulator
from orquestra.opt.optimizers import ScipyOptimizer
from orquestra.opt.problems.maxcut import MaxCut
from orquestra.quantum.estimation import calculate_exact_expectation_values
from orquestra.vqa.ansatz.qaoa_farhi import QAOAFarhiAnsatz
from orquestra.vqa.cost_function.cost_function import (
    create_cost_function,
    substitution_based_estimation_tasks_factory,
)

from icecream import ic


def create_simple_graph():
    """
    [0] --10-- [1]
    |           |
    10          1
    |           |
    [3] --1 -- [2]
    """
    graph = nx.Graph()
    graph.add_nodes_from([0, 1, 2, 3])
    graph.add_edge(0, 1, weight=10)
    graph.add_edge(0, 3, weight=10)
    graph.add_edge(1, 2, weight=1)
    graph.add_edge(2, 3, weight=1)
    return graph


def create_harder_graph():
    graph = nx.Graph()
    graph.add_nodes_from([0, 1, 2, 3, 4, 5])
    graph.add_edge(0, 1, weight=10)
    graph.add_edge(0, 3, weight=5)
    graph.add_edge(1, 2, weight=1)
    graph.add_edge(2, 3, weight=1)

    # YOUR CODE HERE

    graph.add_edge(0, 4, weight=10)
    graph.add_edge(0, 5, weight=5)
    graph.add_edge(1, 4, weight=1)
    graph.add_edge(2, 4, weight=5)
    graph.add_edge(1, 5, weight=5)
    return graph


def solve_maxcut_qaoa(test_graph):

    hamiltonian = MaxCut().get_hamiltonian(test_graph)
    ansatz = QAOAFarhiAnsatz(2, cost_hamiltonian=hamiltonian)

    estimation_method = calculate_exact_expectation_values
    estimation_task_factory = substitution_based_estimation_tasks_factory(
        hamiltonian, ansatz
    )

    backend = CirqSimulator()
    optimizer = ScipyOptimizer(method="L-BFGS-B")

    cost_function = create_cost_function(
        backend,
        estimation_task_factory,
        estimation_method,
    )
    initial_params = np.ones(ansatz.number_of_params) * np.pi / 5
    opt_results = optimizer.minimize(cost_function, initial_params)

    circuit = ansatz.get_executable_circuit(opt_results.opt_params)
    measurements = backend.run_circuit_and_measure(circuit, n_samples=10000)
    counter = Counter(measurements.bitstrings)
    most_common_string = counter.most_common()[0][0]
    return most_common_string


test_graph = create_harder_graph()
most_common_string = solve_maxcut_qaoa(test_graph)
ic(most_common_string)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from functools import partial

import numpy as np
import pytest
from orquestra.integrations.qiskit.simulator import QiskitSimulator
from orquestra.opt.optimizers import ScipyOptimizer
from orquestra.quantum.estimation import estimate_expectation_values_by_averaging
from orquestra.quantum.openfermion import (
    count_qubits,
    get_ground_state,
    get_interaction_operator,
    get_sparse_operator,
    jordan_wigner,
)
from orquestra.quantum.openfermion.hamiltonians import fermi_hubbard
from orquestra.vqa.ansatz.quantum_compiling import HEAQuantumCompilingAnsatz
from orquestra.vqa.cost_function.cost_function import (
    create_cost_function,
    substitution_based_estimation_tasks_factory,
)
from orquestra.vqa.estimation.context_selection import perform_context_selection
from orquestra.vqa.grouping import group_greedily
from orquestra.vqa.shot_allocation import allocate_shots_proportionally


def get_fh_hamiltonian(x_dimension, y_dimension, U):
    hamiltonian = fermi_hubbard(
        x_dimension=x_dimension,
        y_dimension=y_dimension,
        tunneling=1,
        coulomb=U,
        chemical_potential=U / 2,
        magnetic_field=0,
        periodic=False,
        spinless=False,
        particle_hole_symmetry=False,
    )
    operator = get_interaction_operator(hamiltonian)
    operator.one_body_tensor = np.real(operator.one_body_tensor)
    operator.two_body_tensor = np.real(operator.two_body_tensor)
    return operator


class TestVQE:
    def test_solve_fh_with_vqe(self):
        hamiltonian = get_fh_hamiltonian(x_dimension=2, y_dimension=1, U=0.5)
        sparse_operator = get_sparse_operator(hamiltonian)
        exact_energy, _ = get_ground_state(sparse_operator)

        jw_hamiltonian = jordan_wigner(operator=hamiltonian)
        backend = QiskitSimulator("aer_simulator_statevector")

        ansatz = HEAQuantumCompilingAnsatz(2, count_qubits(jw_hamiltonian))
        estimation_method = estimate_expectation_values_by_averaging

        shot_allocation = partial(allocate_shots_proportionally, total_n_shots=50000)
        estimation_preprocessors = [
            group_greedily,
            perform_context_selection,
            shot_allocation,
        ]
        estimation_task_factory = substitution_based_estimation_tasks_factory(
            jw_hamiltonian, ansatz, estimation_preprocessors
        )

        optimizer = ScipyOptimizer(method="COBYLA", options={"maxiter": 250})

        cost_function = create_cost_function(
            backend, estimation_task_factory, estimation_method
        )

        initial_params = np.array(
            [
                0.94557831,
                -1.06991166,
                0.28233565,
                0.47443967,
                0.468148,
                0.5421605,
                0.29240645,
                1.48504768,
                0.48147571,
                0.36578263,
                0.45862184,
                -1.32968834,
                -0.307463,
                0.37939154,
                1.45632776,
                0.09292365,
                0.80710492,
                1.14496005,
                0.65499076,
                -0.06335737,
                0.10678299,
                0.28446654,
                0.5538357,
                0.44774159,
                0.84030601,
                1.77205874,
                0.17598451,
                0.88785342,
                -0.5026344,
                0.72818643,
                -0.00207843,
                0.69474356,
                0.81166782,
                -0.11638733,
                -0.24970387,
                1.02757601,
                0.24950334,
                0.95664434,
                -0.35552715,
                -0.42096424,
                1.30391856,
                0.94936243,
                0.21281127,
                -1.36481705,
                0.37734196,
                0.0475623,
                1.1587143,
                0.9340059,
            ]
        )
        opt_results = optimizer.minimize(cost_function.function, initial_params)
        assert opt_results.opt_value == pytest.approx(exact_energy, rel=2e-2, abs=2e-2)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################

# build the circuit
from orquestra.quantum.circuits import CNOT, H, Circuit
from icecream import ic

bell_circuit = Circuit()
bell_circuit += H(0)
bell_circuit += CNOT(0, 1)
ic(bell_circuit)

bell_circuit2 = Circuit([H(0), CNOT(0, 1)])
ic(bell_circuit2)


# run the circuit
from orquestra.integrations.qiskit.simulator import QiskitSimulator

simulator = QiskitSimulator("aer_simulator")
num_samples = 100
measurements = simulator.run_circuit_and_measure(bell_circuit, num_samples)
ic(measurements.get_counts())

from orquestra.quantum.symbolic_simulator import SymbolicSimulator

sym_simulator = SymbolicSimulator()
measurements2 = sym_simulator.run_circuit_and_measure(bell_circuit, num_samples)
ic(measurements2.get_counts())

sv_simulator = QiskitSimulator("aer_simulator_statevector")
wavefunction = sv_simulator.get_wavefunction(bell_circuit)
ic(wavefunction.amplitudes)

from orquestra.quantum.api.estimation import EstimationTask
from orquestra.quantum.openfermion import IsingOperator
from orquestra.quantum.estimation import calculate_exact_expectation_values

ising = IsingOperator("[Z0] + [Z1]")
task = EstimationTask(ising, bell_circuit, None)
evals = calculate_exact_expectation_values(sym_simulator, [task])
ic(evals[0].values)


# translate the circuit
from orquestra.integrations.qiskit.conversions import (
    export_to_qiskit,
    import_from_qiskit,
)

qiskit_circuit = export_to_qiskit(bell_circuit)
qiskit_circuit.x(1)

from orquestra.integrations.cirq.conversions._circuit_conversions import export_to_cirq

bell_circuit_X = import_from_qiskit(qiskit_circuit)
cirq_circuit = export_to_cirq(bell_circuit_X)
print(cirq_circuit)

wavefunction = sv_simulator.get_wavefunction(bell_circuit_X)
ic(wavefunction.amplitudes)

from orquestra.integrations.forest.conversions import export_to_pyquil

pyquil_circuit = export_to_pyquil(bell_circuit_X)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from collections import Counter

import networkx as nx
import numpy as np
import pytest
from orquestra.integrations.cirq.simulator import CirqSimulator
from orquestra.opt.optimizers import ScipyOptimizer
from orquestra.opt.problems.maxcut import MaxCut
from orquestra.quantum.estimation import calculate_exact_expectation_values
from orquestra.vqa.ansatz.qaoa_farhi import QAOAFarhiAnsatz
from orquestra.vqa.cost_function.cost_function import (
    create_cost_function,
    substitution_based_estimation_tasks_factory,
)


@pytest.fixture
def test_graph():
    """
    [0] --10-- [1]
    |           |
    10          1
    |           |
    [3] --1 -- [2]
    """
    graph = nx.Graph()
    graph.add_nodes_from([0, 1, 2, 3])
    graph.add_edge(0, 1, weight=10)
    graph.add_edge(0, 3, weight=10)
    graph.add_edge(1, 2, weight=1)
    graph.add_edge(2, 3, weight=1)
    return graph


class TestMaxcut:
    def test_solve_maxcut_qaoa(self, test_graph):

        hamiltonian = MaxCut().get_hamiltonian(test_graph)

        ansatz = QAOAFarhiAnsatz(2, cost_hamiltonian=hamiltonian)
        backend = CirqSimulator()
        # 1000 and 0111 are not really optimal solutions, but they're close enough,
        # so that optimization might also end up in this local minimum,
        # and hence it's ok for us to accept it.
        correct_solutions = [
            (1, 0, 1, 0),  # Cost: 22
            (0, 1, 0, 1),  # Cost: 22
            (1, 0, 0, 0),  # Cost: 20
            (0, 1, 1, 1),  # Cost: 20
        ]

        estimation_method = calculate_exact_expectation_values
        estimation_task_factory = substitution_based_estimation_tasks_factory(
            hamiltonian, ansatz
        )

        optimizer = ScipyOptimizer(method="L-BFGS-B")

        cost_function = create_cost_function(
            backend,
            estimation_task_factory,
            estimation_method,
        )
        initial_params = np.ones(ansatz.number_of_params) * np.pi / 5

        opt_results = optimizer.minimize(cost_function, initial_params)
        circuit = ansatz.get_executable_circuit(opt_results.opt_params)
        measurements = backend.run_circuit_and_measure(circuit, n_samples=10000)

        counter = Counter(measurements.bitstrings)
        most_common_string = counter.most_common()[0][0]
        assert most_common_string in correct_solutions



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################

from heapq import nlargest
from numbers import Number
from typing import Callable, cast

import numpy as np
import pytest
from orquestra.integrations.cirq.simulator import CirqSimulator
from orquestra.opt.api import Optimizer
from orquestra.opt.optimizers import CMAESOptimizer
from orquestra.quantum.api.backend import QuantumSimulator
from orquestra.quantum.distributions import (
    MeasurementOutcomeDistribution,
    compute_clipped_negative_log_likelihood,
)
from orquestra.quantum.distributions.BAS_dataset import (
    get_bars_and_stripes_target_distribution,
)
from orquestra.vqa.ansatz.qcbm import QCBMAnsatz
from orquestra.vqa.cost_function.qcbm_cost_function import create_QCBM_cost_function


class TestQCBM:
    @pytest.fixture
    def backend(self):
        return CirqSimulator()

    @pytest.fixture
    def qcbm_ansatz(self):
        return QCBMAnsatz(number_of_layers=4, number_of_qubits=4, topology="all")

    @pytest.fixture
    def target_distribution(self):
        return get_bars_and_stripes_target_distribution(2, 2, 1.0, "zigzag")

    @pytest.fixture
    def optimizer(self):
        return CMAESOptimizer(
            sigma_0=0.1, options={"popsize": 5, "tolx": 1e-6, "seed": 9}
        )

    def test_qcbm_ansatz_optimizes_properly(
        self,
        qcbm_ansatz: QCBMAnsatz,
        backend: QuantumSimulator,
        target_distribution: MeasurementOutcomeDistribution,
        optimizer: Optimizer,
    ):

        cost_function = create_QCBM_cost_function(
            ansatz=qcbm_ansatz,
            backend=backend,
            n_samples=1000,
            distance_measure=cast(
                Callable[..., Number], compute_clipped_negative_log_likelihood
            ),
            distance_measure_parameters={"epsilon": 1e-6},
            target_distribution=target_distribution,
        )

        np.random.seed(9)
        initial_params = np.random.uniform(-1.57, 1.57, qcbm_ansatz.number_of_params)

        opt_results = optimizer.minimize(cost_function, initial_params, True)

        actual_distribution = backend.get_measurement_outcome_distribution(
            qcbm_ansatz.get_executable_circuit(opt_results.opt_params)
        )

        size_of_bar_and_stripes = len(target_distribution.distribution_dict)
        assert set(
            nlargest(
                size_of_bar_and_stripes,
                actual_distribution.distribution_dict,
                key=actual_distribution.distribution_dict.get,  # type: ignore
            )
        ) == set(target_distribution.distribution_dict.keys())
        assert opt_results.opt_value < 1.8
